from glyles.glycans.factory.factory_f import FuranoseFactory
from glyles.glycans.factory.factory_p import PyranoseFactory
from glyles.glycans.nx_monomer import NXMonomer
from glyles.glycans.rdkit_monomer import RDKitMonomer
from glyles.glycans.utils import Mode, UnreachableError
from glyles.grammar.GlycanLexer import GlycanLexer


class MonomerFactory:
    """
    Class holding and managing the access and generation of all monomers implemented in this package.
    """

    def __init__(self):
        """
        Initialize this factory by creating instances of all "sub-"factories
        """
        self._pyranoses = PyranoseFactory()
        self._furanoses = FuranoseFactory()

        self._keys = set(self._pyranoses.keys())

    def __contains__(self, item):
        """
        Check if an item is part of this factory and can be returned.

        Args:
            item (str): Code of a monomer to be checked

        Returns:
            True if the item is included in the current version of this package
        """
        return item.upper() in self._keys

    def __getitem__(self, item):
        """
        Get an instance of a monomer from this factory.

        Args:
            item (str): name of the query monomer

        Returns:
            Directory containing all necessary information to initialize a monomer implementation
        """
        furanose = False
        if item[-1] == "p" and not item.endswith("manHep"):
            item = item[:-1]
        if item[-1] == "f":
            item = item[:-1]
            furanose = True

        if furanose and item in self._furanoses:
            return self._furanoses[item]
        if not furanose and item in self._pyranoses:
            return self._pyranoses[item]
        raise NotImplementedError("Query-monomer is neither in pyranoses nor in furanoses")

    def keys(self):
        """
        Get all monomers that are included in this package by their extended name.

        Returns:
            Set of names for all monomers, their available derivatives and configurations (alpha/beta/undefined)
        """
        return self._keys

    def monomers(self):
        """
        Get the names of all monomers in this package ignoring the alpha/beta conformations.

        Returns:
            List, sorted from long to short, of all monomer names in upper case
        """
        return sorted(set([self[x.split("_")[-1]]["name"] for x in self._keys]), key=lambda x: -len(x))

    def furanoses(self):
        """
        Get the names of all monomers available as furanoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all furanose names in upper case
        """
        return sorted(set([self[x.split("_")[-1]]["name"] for x in self._furanoses.keys()]))

    def pyranoses(self):
        """
        Get the names of all monomers available as pyranoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all pyranose names in upper case
        """
        return sorted(set([self[x.split("_")[-1]]["name"] for x in self._pyranoses.keys()]))

    def monomer_names(self):
        """
        Get the names of all monomers in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all monomer names in IUPAC notation
        """
        output = set()
        for item in self.monomers():
            if item in self._pyranoses:
                output.add(self._pyranoses[item]["name"])
        return list(output)

    def pyranose_names(self):
        """
        Get the names of all pyranoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all monomer names in IUPAC notation
        """
        output = set()
        for item in self.pyranoses():
            if item in self._pyranoses:
                output.add(self._pyranoses[item]["name"])
        return list(output)

    def furanose_names(self):
        """
        Get the names of all furanoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all monomer names in IUPAC notation
        """
        output = set()
        for item in self.monomers():
            if item in self._furanoses:
                output.add(self._furanoses[item]["name"])
        return list(output)

    def create(self, recipe, mode=Mode.RDKIT_MODE, config=None):
        """
        Create a monomer from its describing IUPAC string with all added side chains.

        Args:
            recipe (List[Tuple[str, int]]): List of modifications, conformations and the root monomer
            mode (Mode): implementation used to represent monomers
            config (str): configuration if monomer is alpha or beta monomer

        Returns:
            Monomer-Instance containing all modifications given in the input
        """
        # determine the class to use to represent the monomers
        if mode == Mode.NETWORKX_MODE:
            monomer_class = NXMonomer
        elif mode == Mode.RDKIT_MODE:
            monomer_class = RDKitMonomer
        else:
            raise ValueError("Unknown representation mode for monomers!")

        # extract key information from the input, i.e. the type, the configuration and pyranose/furanose
        tmp = list(zip(*recipe))
        name = recipe[tmp[1].index(GlycanLexer.SAC)][0]
        config_index = tmp[1].index(GlycanLexer.TYPE) if GlycanLexer.TYPE in tmp[1] else None
        ring_index = tmp[1].index(GlycanLexer.RING) if GlycanLexer.RING in tmp[1] else None

        # generate the full name that is looked up in the factory
        if config is not None and len(config) > 0:
            name = config + "_" + name
        elif config_index is not None:
            name = recipe[config_index][0] + "_" + name

        # get the monomer from the factory
        try:
            if ring_index is not None and recipe[ring_index][0] == "f":
                monomer = monomer_class(**self._furanoses[name], recipe=recipe)
            else:
                monomer = monomer_class(**self._pyranoses[name], recipe=recipe)
        except KeyError:
            raise UnreachableError("Invalid monomer found, should never get so far.")

        # create the final molecule using the molecule's react-method augmented with the recipe of the molecule
        monomer = monomer.react(*tmp)

        return monomer



import os
import sys

from glyles.glycans.factory.factory import MonomerFactory
from glyles.grammar.parse import Glycan


def preprocess_glycans(glycan, glycan_list, glycan_file):
    """
    Preprocess the static inputs for the parsing into one single list

    Args:
        glycan (str): single glycan to parse
        glycan_list (List[str]): list of glycans to parse
        glycan_file (str): filepath of file to read glycans from

    Returns:
        list of glycans in the order they are handed to the function, i.e. glycan, glycan_list, glycan_file
    """
    glycans = []

    # fill a list with all glycans to convert
    if glycan is not None:
        glycans.append(glycan)
    if glycan_list is not None:
        glycans += glycan_list
    if glycan_file is not None:
        # check if the file is valid and read it out
        if not os.path.isfile(glycan_file):
            pass
        for line in open(glycan_file, "r").readlines():
            glycans.append(line.strip())
    return glycans


def convert(glycan=None, glycan_list=None, glycan_file=None, glycan_generator=None, output_file=None, returning=False,
            silent=True):
    """
    General user interaction interface to use this library.

    Args:
        glycan (str): Single glycan to be converted from IUPAC to SMILES
        glycan_list (List[str]): list of glycans to convert
        glycan_file (str): File to read the glycans from
        glycan_generator (generator): generator yielding iupac representation.
            Together with output_generator=True this does not create any lists
        output_file (str): File to save the converted glycans in
        returning (bool): Flag indicating to return a list of tuples
        silent (bool): Flag indicating to have no prints from this method

    Returns:
        List of type (IUPAC, SMILES) items giving the converted SMILES formulas. Only if returning=True is set.
    """

    # collect all data and return if no data were provided
    glycans = preprocess_glycans(glycan, glycan_list, glycan_file)
    if len(glycans) == 0 and glycan_generator is None:
        if not silent:
            print("List of glycans is empty", file=sys.stderr)
        return

    # determine the output format
    if output_file is not None:
        if os.path.isdir(os.path.dirname(os.path.abspath(output_file))):
            output = open(output_file, "w")
        else:
            if not silent:
                print("Path of output-file does not exist! Results will be printed on stdout.", file=sys.stderr)
            output = sys.stdout
    else:
        if returning:
            output = []
        else:
            if not silent:
                print("No output-file specified, results will be printed on stdout.")
            output = sys.stdout

    # convert the IUPAC strings into SMILES strings from the input list
    if len(glycans) != 0:
        for glycan, smiles in convert_generator(glycan_list=glycans):
            if returning:
                output.append((glycan, smiles))
            else:
                print(glycan, smiles, file=output, sep=",")

    # and from the input generator
    if glycan_generator is not None:
        for glycan, smiles in convert_generator(glycan_generator=glycan_generator):
            if returning:
                output.append((glycan, smiles))
            else:
                print(glycan, smiles, file=output, sep=",")

    if returning:
        return output
    elif output_file is not None:
        output.close()


def convert_generator(glycan=None, glycan_list=None, glycan_file=None, glycan_generator=None, silent=True):
    """
    General user interaction interface to use this library.

    Args:
        glycan (str): Single glycan to be converted from IUPAC to SMILES
        glycan_list (List[str]): list of glycans to convert
        glycan_file (str): File to read the glycans from
        glycan_generator (generator): generator yielding iupac representation.
            Together with output_generator=True this does not create any lists
        silent (bool): Flag indicating to have no output-messages from this method

    Returns:
        Generator generating pairs of type (IUPAC, SMILES) items giving the converted SMILES formulas for the IUPACs.
    """
    factory = MonomerFactory()
    glycans = preprocess_glycans(glycan, glycan_list, glycan_file)
    if len(glycans) == 0 and glycan_generator is None:
        if not silent:
            print("List of glycans is empty", file=sys.stderr)
        return

    # Convert the glycans ...
    if len(glycans) != 0:
        for glycan in glycans:
            try:
                # ... by passing them to the glycan class to parse them and return them as intended
                yield glycan, Glycan(glycan, factory).get_smiles()

            # catch any exception at glycan level to not destroy the whole pipeline because of one mis-formed glycan
            except Exception as e:
                print(f"An exception occurred with {glycan}:", e.__class__, file=sys.stderr)
                print("Error message:", e.__str__(), file=sys.stderr)
                yield glycan, ""

    # Convert the glycans ...
    if glycan_generator is not None:
        for glycan in glycan_generator:
            try:
                # ... by passing them to the glycan class to parse them and return them as intended
                yield glycan, Glycan(glycan, factory).get_smiles()

            # catch any exception at glycan level to not destroy the whole pipeline because of one mis-formed glycan
            except Exception as e:
                print(f"An exception occurred with {glycan}:", e.__class__, file=sys.stderr)
                print("Error message:", e.__str__(), file=sys.stderr)
                yield glycan, ""



import sys

import networkx as nx
import numpy as np
import pydot
from antlr4 import *
from rdkit.Chem import MolFromSmiles

from glyles.glycans.utils import Mode, UnreachableError, ParseError
from glyles.grammar.GlycanLexer import GlycanLexer
from glyles.grammar.GlycanParser import GlycanParser


class Glycan:
    """
    This class is like an interaction with the Parser for the IUPAC representation of the glycan. The grammar for
    glycans is defined using ANTLR (https://www.antlr.org/). From this ANTLR is able to generate lexer and parser that
    fit the defined grammar. Don't touch those files those are auto generated and therefore mostly uncommented.

    The defined grammar discards the last glycan which is used to define the root of the glycan tree. Therefore, the
    resulting abstract syntax trees (AST)s are not intuitive.
    """

    class __TreeWalker:
        def __init__(self, factory):
            """
            Just initialize an empty tree

            Args:
                factory (MonomerFactory): factory instance to use to generate the monomers for the glycan tree from
            """
            self.g = nx.DiGraph()
            self.factory = factory
            self.node_id = 0

        def parse(self, t, mode):
            """
            Parse a parsed tree (AST) from ANTLR into this networkx graph

            Args:
                t (antlr.ParseTree): result of the parsing step from ANTLR
                mode (Mode): mode determining which monomer-mode to use

            Returns:
                Tree of parsed glycan with monomers in nodes
            """

            # parse the initial monomer and the orientation of the root monomer
            # and remove the first and last char, i.e. '{' and '}'
            children = list(t.getChildren())[1:-1]
            if len(children) == 1:  # glycan
                self.__add_node(children[0], mode)
                return self.g
            elif len(children) == 2:  # branch glycan
                node_id = self.__add_node(children[1], mode)
                self.__walk(children[0], node_id, mode)
                return self.g
            elif len(children) == 3:  # SAC ' ' TYPE
                self.__add_node(children[0], mode, children[2].symbol.text)
                return self.g
            elif len(children) == 4:  # branch SAC ' ' TYPE
                node_id = self.__add_node(children[1], mode, children[3].symbol.text)
                self.__walk(children[0], node_id, mode)
                return self.g
            else:
                raise RuntimeError("This branch of the if-statement should be unreachable!")

        def __walk(self, t, parent, mode):
            """
            The heart of this class. This method recursively adds nodes to the graph and connects them according to the
            connections in the input IUPAC.

            Args:
                t (antlr.ParseTree): subtree to parse in this recursive step.
                parent (int): ID of the parent node for this (subtree)
                mode (Mode): mode determining which monomer-mode to use

            Returns:
                ID of the parent for the remaining recursive procedure after resolving this subtree
            """

            # security check for some nodes that should not occur, but who knows what ANTLR does (or does not) ;-)
            if isinstance(t, ErrorNode):
                raise UnreachableError("ErrorNodes in parsing!")
            elif isinstance(t, TerminalNode):
                raise UnreachableError("TerminalNodes in parsing!")

            children = list(t.getChildren())
            if len(children) == 2:  # {glycan con}
                # terminal element, add the node with the connection
                node_id = self.__add_node(children[0], mode)
                self.__add_edge(parent, node_id, children[1])
                return node_id

            elif len(children) == 3 and isinstance(children[2], GlycanParser.BranchContext):  # {glycan con branch}
                # chain without branching, the parent is the parent of the parsing of the back part
                parent = self.__walk(children[2], parent, mode)
                node_id = self.__add_node(children[0], mode)
                self.__add_edge(parent, node_id, children[1])
                return node_id

            elif len(children) == 3 and isinstance(children[1], GlycanParser.BranchContext):  # {'[' branch ']'}
                # branching, hand the parent on to the next level
                self.__walk(children[1], parent, mode)
                return parent

            elif len(children) == 6:  # {glycan con '[' branch ']' branch}
                # branching in a chain, append the end to the parent and hang both branches on that
                node_id = self.__walk(children[5], parent, mode)
                self.__walk(children[3], node_id, mode)
                node_id2 = self.__add_node(children[0], mode)
                self.__add_edge(node_id, node_id2, children[1])
                return node_id2

            # there should be no case missing, but who knows...
            raise UnreachableError("Invalid branching in glycan tree")

        def __add_edge(self, parent, child, con):
            """
            Add an edge between the provided ids of the parent and the children in the glycan tree.

            Args:
                parent (int): ID of parent in the connection of the glycan tree
                child (int): ID of the child in the connection of the glycan tree
                con (str): Connection element from the ParseTree

            Returns:
                Nothing
            """
            self.g.add_edge(parent, child, type=str(con))

        def __add_node(self, node, mode, config=""):
            """
            Add a new node to the network based on the name of the represented glycan.

            Args:
                node: Node of the parsed tree to be parsed into a monomer
                mode (Mode): mode determining which monomer-mode to use
                config (str): configuration to be applied to the monomer

            Returns:
                ID of the newly added node
            """
            # get the id for the node and increase the id for the next node
            node_id = self.node_id
            self.node_id += 1

            # add the node to the network and store the enum of the glycan as attribute
            recipe = []
            for child in node.children:
                if isinstance(child, GlycanParser.DerivContext):
                    for c in child.children:
                        recipe.append((str(c), c.symbol.type))
                else:
                    recipe.append((str(child), child.symbol.type))
            self.g.add_node(
                node_id,
                type=self.factory.create(recipe, mode, config),
            )

            return node_id

    class __Merger:
        """
        Merge the tree of monomers into a SMILES representation of the complete molecule.
        """

        def __init__(self, factory):
            """
            Create a merger class to merge a parsed glycan tree into a SMILES string

            Args:
                factory (MonomerFactory): factory instance to use to generate the monomers for the glycan tree from
            """
            self.factory = factory

        def merge(self, t, root_orientation="n", start=10):
            """
            Merge the provided tree of monomers enriched with the glycans in the nodes and information on the bindings
            between two monomer-nodes in the edges.

            Args:
                t (networkx.DiGraph): Graph representing the glycan to compute the whole SMILES representation for.
                root_orientation (str): Orientation of the root monomer in glycan (can be a (= alpha) or b (= beta))
                start (int): index of the atom to start the SMILES generation from

            Returns:
                SMILES representation as string
            """
            # first mark the atoms that will be replaced in a binding of two monomers
            self.__mark(t, 0, "({}1-?)".format(root_orientation))

            # return the string that can be computed from connecting the monomers as marked above
            position = int(np.argwhere(t.nodes[0]["type"].get_features()[:, 1] == start).squeeze())
            return self.__merge(t, 0, position, 0)

        def __mark(self, t, node, p_edge):
            """
            Recursively mark in every node of the molecule which atoms are being replaced by bound monomers.

            Args:
                t (networkx.DiGraph): Tree representing the glycan to compute the whole SMILES representation for.
                node (int): ID of the node to work on in this method
                p_edge (str): edge annotation to parent monomer

            Returns:
                Nothing
            """
            # get children nodes
            children = [x[1] for x in t.edges(node)]

            # set chirality of atom binding parent
            if p_edge is not None and t.nodes[node]["type"].is_non_chiral():
                t.nodes[node]["type"] = t.nodes[node]["type"].to_chirality(p_edge[1], self.factory)

            # check for validity of the tree, ie if it's a leaf (return, nothing to do) or has too many children (Error)
            if len(children) == 0:  # leaf
                return
            if len(children) > 2:  # too many children
                raise NotImplementedError("Glycans with maximal branching factor greater then 2 not implemented.")

            # iterate over the children and the atoms used to mark binding atoms in my structure
            for child, atom in zip(children, t.nodes[node]["type"].get_dummy_atoms()[0]):
                binding = t.get_edge_data(node, child)["type"]

                t.nodes[node]["type"].mark(int(binding[4]), atom)
                self.__mark(t, child, binding)

        def __merge(self, t, node, start, ring_index):
            """
            Recursively merge every node of the molecule with its children and get the SMILES representation of the
            subtree.

            Args:
                t (networkx.DiGraph): Graph representing the glycan to compute the whole SMILES representation for.
                node (int): ID of the node to work on in this method
                start (int): ID of the atom in the inner graph to start from when generating the SMILES string
                ring_index (int): Index of the ring to use when generating the SMILES strings

            Returns:
                SMILES representation of the subtree of the given node
            """
            # get my children and compute my SMILES string
            children = [x[1] for x in t.edges(node)]
            me = t.nodes[node]["type"].to_smiles(start, ring_index)

            # check for validity of the tree, ie if it's a leaf
            if len(children) == 0:  # leaf
                return me

            # iterate over the children and the atoms used to mark binding atoms
            for child, atom in zip(children, t.nodes[node]["type"].get_dummy_atoms()[1]):
                binding = list(t.get_edge_data(node, child)["type"])

                # child_start = t.nodes[node]["type"].root_atom_id(int(binding[2]))
                child_start = t.nodes[child]["type"].root_atom_id(int(binding[2]))
                if child_start == -1:
                    raise ValueError("No child start found.")

                # get the SMILES of this child and plug it in the current own SMILES
                child_smiles = self.__merge(t, child, child_start, ring_index + 1)
                me = me.replace(atom, child_smiles)
            return me

    def __init__(self, iupac, factory, mode=Mode.DEFAULT_MODE, root_orientation="n", start=10, parse=True):
        """
        Initialize the glycan from the IUPAC string.

        Args:
            iupac (str): IUPAC string representation of the glycan to represent
            factory (MonomerFactory): factory instance to use to generate the monomers for the glycan tree from
            mode (Mode): Glycan mode to use. This controls the representation used (either networkx or RDKit)
            root_orientation (str): orientation of the root monomer in the glycan (choose from 'a', 'b', 'n')
            start (int): ID of the atom to start with in the root monomer when generating the SMILES
            parse (bool): Flag indicating to also parse the SMILES immediately
        """
        self.iupac = iupac
        self.mode = mode
        self.parse_tree = None
        self.glycan_smiles = None
        self.root_orientation = root_orientation
        self.start = start
        self.parse_smiles = parse
        self.factory = factory
        self.__parse()

    def get_smiles(self):
        """
        Request the SMILES string of the parsed molecule.

        Returns:
            Generated SMILES string
        """
        if self.glycan_smiles is None:
            self.glycan_smiles = Glycan.__Merger(self.factory).merge(self.parse_tree, self.root_orientation,
                                                                     start=self.start)
        self.glycan_smiles = self.glycan_smiles.replace("At", "O-")
        return self.glycan_smiles

    def get_tree(self):
        """
        Request the tree parsed from the IUPAC in this instance.

        Returns:
            The parsed tree with the single monomers in the nodes
        """
        return self.parse_tree

    def to_pdb(self, output):
        """
        Compute a 3-dimensional conformation of the molecule using ForceFields. The result will be stored in a PDB file.

        Args:
            output (str): Path to store the PDB file in

        Returns:
            Nothing
        """
        mol = MolFromSmiles(self.glycan_smiles)
        raise NotImplementedError("PDB conversion planned, but not implemented yet.")

    def save_dot(self, output):
        """
        Save the tree structure of the encoded glycan molecule into a dot file visualizing the graph of monomers.

        Args:
            output (str): path to store the DOT file in

        Returns:
            Nothing
        """
        graph = pydot.Dot("iupac_tree")
        for node in range(len(self.parse_tree.nodes)):
            graph.add_node(pydot.Node(node, label=self.parse_tree.nodes[node]["type"].get_name()))
        for edge in self.parse_tree.edges():
            graph.add_edge(pydot.Edge(*edge, label=self.parse_tree.get_edge_data(*edge)["type"]))
        graph.write(output + ".dot")

    def __parse(self):
        """
        Adapter on the Lexer and Parser generated by ANTLR based on Grammar.g4.

        Returns:
            Nothing
        """
        # catch the prints of antlr to stderr to check if during parsing an error occurred and the glycan is invalid
        log = []

        class writer(object):
            @staticmethod
            def write(data):
                log.append(data)

        old_err = sys.stderr
        sys.stderr = writer()

        # parse the remaining structure description following the grammar, also add the dummy characters
        stream = InputStream(data='{' + self.iupac + '}')
        lexer = GlycanLexer(stream)
        token = CommonTokenStream(lexer)
        parser = GlycanParser(token)
        tree = parser.start()

        sys.stderr = old_err

        # if the glycan is invalid, set its structure to None and the SMILES string to empty and return
        if len(log) != 0:
            self.parse_tree = None
            self.glycan_smiles = ""
            raise ParseError("Glycan cannot be parsed:\n" + log[0])

        # walk through the AST and parse the AST into a networkx representation of the glycan.
        self.parse_tree = Glycan.__TreeWalker(self.factory).parse(tree, self.mode)

        # if the glycan should be parsed immediately, do so
        if self.parse_smiles:
            self.glycan_smiles = Glycan.__Merger(self.factory).merge(self.parse_tree, self.root_orientation,
                                                                     start=self.start)



from enum import Enum


class Config(Enum):
    """
    Configuration enum to represent if the monomer is in alpha, beta, or undefined configuration.
    """
    UNDEF = 0
    ALPHA = 1
    BETA = 2


class Enantiomer(Enum):
    """
    Configuration of the whole monomer regarding L and D forms in terms of enantiomerism.
    """
    D = 0
    L = 1


class Lactole(Enum):
    """
    Specification if a monomer is a pyranose (6-ring) or a furanose (5-ring)
    """
    FURANOSE = 5
    PYRANOSE = 6


class Mode(Enum):
    """
    Enumerate different modes how to represent the monomers in the tree.
    """
    DEFAULT_MODE = "rdkit"
    NETWORKX_MODE = "nx"
    RDKIT_MODE = "rdkit"


class UnreachableError(NotImplementedError):
    """
    Represent exceptions that should arise in case a piece of code is reached that under normal circumstances should
    not be reached
    """
    pass


class ParseError(ValueError):
    """
    Represent parsing errors when reading in a glycan
    """
    pass


class Node:
    """
    Represent a node in a tree
    """

    def __init__(self, node_id, parent_id, depth, tree):
        """
        Initialize a node in of a tree
        Args:
            node_id (int): id of the node itself in the tree
            parent_id (int): id of the parent in the tree
            depth (int): depth of the node in the tree this node belongs to
            tree (Tree): reference to the tree this node belongs to
        """
        self.children = []
        self.node_id = node_id
        self.parent_id = parent_id
        self.depth = depth
        self.tree = tree

    def add_child(self, child_id):
        """
        Add a child node to this node

        Args:
            child_id (int): id of the added child node

        Returns:
            Nothing
        """
        self.children.append(child_id)

    def is_leaf(self):
        """
        Check if this node is a leaf

        Returns:
            true if this node has no children
        """
        return len(self.children) == 0

    def __str__(self):
        """
        Convert the node into a string representation

        Returns:
            String representation of this node and all its children
        """
        return "(" + str(self.node_id) + " [" + \
               ",".join([str(self.tree.nodes[child]) for child in self.children]) + "])"


class Tree:
    """
    Represent a tree with nodes
    """

    def __init__(self):
        """
        Initialize the tree as empty tree without any node
        """
        self.nodes = {}
        self.root = None

    def __str__(self):
        """
        Convert this tree into a string representation of all its nodes

        Returns:
            String representation of the tree seen from its root node
        """
        return str(self.nodes[self.root])

    def add_node(self, node_id, parent_id=-1):
        """
        Add a node to the tree

        Args:
            node_id (int): id of the new node
            parent_id (int): id of the parent of the new node

        Returns:
            Nothing
        """
        # check if the node is a root and the tree already has a root
        if parent_id == -1 and self.root is not None:
            raise ValueError("Tree cannot have two roots")
        if parent_id != -1 and parent_id not in self.nodes:
            raise ValueError("Parent node unknown")

        # add the node and eventually set the root to be that node
        if parent_id == -1:
            self.root = node_id
            self.nodes[node_id] = Node(node_id, parent_id, 0, self)
        else:
            self.nodes[node_id] = Node(node_id, parent_id, self.nodes[parent_id].depth + 1, self)

        # add the node to the list of children in its parent
        if parent_id != -1:
            self.nodes[parent_id].add_child(node_id)

    def deepest_node(self):
        """
        Find the deepest node in the tree (most distant to root)

        Returns:
            ID of the deepest node and its depth in the tree
        """
        deepest_id, deepest_depth = 0, 0
        for n_id, node in self.nodes.items():
            if node.depth > deepest_depth:
                deepest_depth = node.depth
                deepest_id = n_id
        return deepest_id, deepest_depth

    def rehang_tree(self, node_id):
        """
        Reorder the tree to start with the node with the given id in the root

        Args:
            node_id (int): is of the node to be used as new root node

        Returns:
            New tree with the specified node as root node
        """
        tree = Tree()
        stack = [(-1, node_id)]
        while len(stack) != 0:
            p_id, c_id = stack[-1]
            stack = stack[:-1]
            tree.add_node(c_id, p_id)

            children = self.nodes[c_id].children + [self.nodes[c_id].parent_id]
            for c in children:
                if c not in tree.nodes and c != -1:
                    stack.append((c_id, c))

        return tree

    def longest_chain(self, node_id=None):
        """
        Find the longest chain of nodes in the parent. This chain starts in the provided node and goes down in the
        tree. This might lead to false results of the node with the given id has two or more children.

        Args:
            node_id (int): id of the node to start in, if None, the root will be the start

        Returns:
            List of nodes along deepest path down the tree
        """
        if node_id is None:
            node_id = self.root

        if self.nodes[node_id].is_leaf():
            return [node_id]

        longest_chain = []
        for child in self.nodes[node_id].children:
            tmp = self.longest_chain(child)
            if len(tmp) > len(longest_chain):
                longest_chain = tmp

        return [node_id] + longest_chain



# Generated from Glycan.g4 by ANTLR 4.9
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .GlycanParser import GlycanParser
else:
    from GlycanParser import GlycanParser

# This class defines a complete listener for a parse tree produced by GlycanParser.
class GlycanListener(ParseTreeListener):

    # Enter a parse tree produced by GlycanParser#start.
    def enterStart(self, ctx:GlycanParser.StartContext):
        pass

    # Exit a parse tree produced by GlycanParser#start.
    def exitStart(self, ctx:GlycanParser.StartContext):
        pass


    # Enter a parse tree produced by GlycanParser#branch.
    def enterBranch(self, ctx:GlycanParser.BranchContext):
        pass

    # Exit a parse tree produced by GlycanParser#branch.
    def exitBranch(self, ctx:GlycanParser.BranchContext):
        pass


    # Enter a parse tree produced by GlycanParser#glycan.
    def enterGlycan(self, ctx:GlycanParser.GlycanContext):
        pass

    # Exit a parse tree produced by GlycanParser#glycan.
    def exitGlycan(self, ctx:GlycanParser.GlycanContext):
        pass


    # Enter a parse tree produced by GlycanParser#deriv.
    def enterDeriv(self, ctx:GlycanParser.DerivContext):
        pass

    # Exit a parse tree produced by GlycanParser#deriv.
    def exitDeriv(self, ctx:GlycanParser.DerivContext):
        pass



del GlycanParser


from glyles.glycans.utils import *


class PyranoseFactory:
    """
    Checked with http://www.cheminfo.org/flavor/malaria/Utilities/SMILES_generator___checker/index.html
    """

    __monomers = {
        # Glucose
        "GLC": {"name": "Glc", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_GLC": {"name": "Glc", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_GLC": {"name": "Glc", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},

        # Mannose
        "MAN": {"name": "Man", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@H]1OC(O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "A_MAN": {"name": "Man", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "B_MAN": {"name": "Man", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},

        # Galactose
        "GAL": {"name": "Gal", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@H]1O"},
        "A_GAL": {"name": "Gal", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@H]1O"},
        "B_GAL": {"name": "Gal", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@H]1O"},

        # Gulose
        "GUL": {"name": "Gul", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@H](O)[C@H]1O"},
        "A_GUL": {"name": "Gul", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@H](O)[C@H]1O"},
        "B_GUL": {"name": "Gul", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@H](O)[C@H]1O"},

        # Altrose
        "ALT": {"name": "Alt", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@@H]1OC(O)[C@H](O)[C@@H](O)[C@H]1O"},
        "A_ALT": {"name": "Alt", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@H]1O"},
        "B_ALT": {"name": "Alt", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@H]1O"},

        # Allose
        "ALL": {"name": "All", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@H](O)[C@@H]1O"},
        "A_ALL": {"name": "All", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@H](O)[C@@H]1O"},
        "B_ALL": {"name": "All", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@H](O)[C@@H]1O"},

        # Talose
        "TAL": {"name": "Tal", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@H]1OC(O)[C@@H](O)[C@@H](O)[C@H]1O"},
        "A_TAL": {"name": "Tal", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@@H](O)[C@@H](O)[C@H]1O"},
        "B_TAL": {"name": "Tal", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@@H](O)[C@@H](O)[C@H]1O"},

        # Idose
        "IDO": {"name": "Ido", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "OC[C@@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_IDO": {"name": "Ido", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_IDO": {"name": "Ido", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},

        # Quinovose
        "QUI": {"name": "Qui", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_QUI": {"name": "Qui", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_QUI": {"name": "Qui", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},

        # Rhamnose
        "RHA": {"name": "Rha", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@@H]1OC(O)[C@H](O)[C@H](O)[C@H]1O"},
        "A_RHA": {"name": "Rha", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H]1O[C@@H](O)[C@H](O)[C@H](O)[C@H]1O"},
        "B_RHA": {"name": "Rha", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H]1O[C@H](O)[C@H](O)[C@H](O)[C@H]1O"},

        # Fucose
        "FUC": {"name": "Fuc", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@@H]1OC(O)[C@@H](O)[C@H](O)[C@@H]1O"},
        "A_FUC": {"name": "Fuc", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H]1O[C@H](O)[C@@H](O)[C@H](O)[C@@H]1O"},
        "B_FUC": {"name": "Fuc", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H]1O[C@@H](O)[C@@H](O)[C@H](O)[C@@H]1O"},

        # Olivose
        "OLI": {"name": "Oli", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)C[C@@H](O)[C@@H]1O"},
        "A_OLI": {"name": "Oli", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)C[C@@H](O)[C@@H]1O"},
        "B_OLI": {"name": "Oli", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)C[C@@H](O)[C@@H]1O"},

        # Tyvelose
        "TYV": {"name": "Tyv", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)[C@@H](O)C[C@@H]1O"},
        "A_TYV": {"name": "Tyv", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)[C@@H](O)C[C@@H]1O"},
        "B_TYV": {"name": "Tyv", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)[C@@H](O)C[C@@H]1O"},

        # Abequose
        "ABE": {"name": "Abe", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)[C@H](O)C[C@H]1O"},
        "A_ABE": {"name": "Abe", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)[C@H](O)C[C@H]1O"},
        "B_ABE": {"name": "Abe", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)[C@H](O)C[C@H]1O"},

        # Paratose
        "PAR": {"name": "Par", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)[C@H](O)C[C@@H]1O"},
        "A_PAR": {"name": "Par", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)[C@H](O)C[C@@H]1O"},
        "B_PAR": {"name": "Par", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)[C@H](O)C[C@@H]1O"},

        # Digitoxose
        "DIG": {"name": "Dig", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)C[C@H](O)[C@@H]1O"},
        "A_DIG": {"name": "Dig", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)C[C@H](O)[C@@H]1O"},
        "B_DIG": {"name": "Dig", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)C[C@H](O)[C@@H]1O"},

        # Colitose
        "COL": {"name": "Col", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@@H]1OC(O)[C@@H](O)C[C@@H]1O"},
        "A_COL": {"name": "Col", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H]1O[C@H](O)[C@@H](O)C[C@@H]1O"},
        "B_COL": {"name": "Col", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H]1O[C@@H](O)[C@@H](O)C[C@@H]1O"},

        # Arabinose
        "ARA": {"name": "Ara", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "O[C@H]1COC(O)[C@H](O)[C@H]1O"},
        "A_ARA": {"name": "Ara", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@H]1CO[C@@H](O)[C@H](O)[C@H]1O"},
        "B_ARA": {"name": "Ara", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@H]1CO[C@H](O)[C@H](O)[C@H]1O"},

        # Lyxose
        "LYX": {"name": "Lyx", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "O[C@@H]1COC(O)[C@@H](O)[C@H]1O"},
        "A_LYX": {"name": "Lyx", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@@H]1CO[C@H](O)[C@@H](O)[C@H]1O"},
        "B_LYX": {"name": "Lyx", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@@H]1CO[C@@H](O)[C@@H](O)[C@H]1O"},

        # Xylose
        "XYL": {"name": "Xyl", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "O[C@@H]1COC(O)[C@H](O)[C@H]1O"},
        "A_XYL": {"name": "Xyl", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@@H]1CO[C@H](O)[C@H](O)[C@H]1O"},
        "B_XYL": {"name": "Xyl", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@@H]1CO[C@@H](O)[C@H](O)[C@H]1O"},

        # Ribose
        "RIB": {"name": "Rib", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "O[C@@H]1COC(O)[C@H](O)[C@@H]1O"},
        "A_RIB": {"name": "Rib", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@@H]1CO[C@H](O)[C@H](O)[C@@H]1O"},
        "B_RIB": {"name": "Rib", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O[C@@H]1CO[C@@H](O)[C@H](O)[C@@H]1O"},

        # Keto-Deoxy-Nonulonic acid
        "KDN": {"name": "Kdn", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "O=C(O)C1(O)C[C@H](O)[C@@H](O)C([C@H](O)[C@H](O)CO)O1"},
        "A_KDN": {"name": "Kdn", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O=C(O)[C@@]1(O)C[C@H](O)[C@@H](O)C([C@H](O)[C@H](O)CO)O1"},
        "B_KDN": {"name": "Kdn", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O=C(O)[C@]1(O)C[C@H](O)[C@@H](O)C([C@H](O)[C@H](O)CO)O1"},

        # Neuraminic acid
        "NEU": {"name": "Neu", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "N[C@@H]1[C@@H](O)CC(O)(C(=O)O)O[C@H]1[C@H](O)[C@H](O)CO"},
        "A_NEU": {"name": "Neu", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "N[C@@H]1[C@@H](O)C[C@](O)(C(=O)O)O[C@H]1[C@H](O)[C@H](O)CO"},
        "B_NEU": {"name": "Neu", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "N[C@@H]1[C@@H](O)C[C@@](O)(C(=O)O)O[C@H]1[C@H](O)[C@H](O)CO"},

        # Pseudaminic acid
        "PSE": {"name": "Pse", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H](O)[C@H](N)C1OC(O)(C(=O)O)C[C@H](O)[C@@H]1N"},
        "A_PSE": {"name": "Pse", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H](O)[C@H](N)C1O[C@@](O)(C(=O)O)C[C@H](O)[C@@H]1N"},
        "B_PSE": {"name": "Pse", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H](O)[C@H](N)C1O[C@](O)(C(=O)O)C[C@H](O)[C@@H]1N"},

        # Legionaminic acid
        "LEG": {"name": "Leg", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@@H](O)[C@@H](N)C1OC(O)(C(=O)O)C[C@H](O)[C@H]1N"},
        "A_LEG": {"name": "Leg", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H](O)[C@@H](N)C1O[C@@](O)(C(=O)O)C[C@H](O)[C@H]1N"},
        "B_LEG": {"name": "Leg", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H](O)[C@@H](N)C1O[C@](O)(C(=O)O)C[C@H](O)[C@H]1N"},

        # Acinetaminic acid
        "ACI": {"name": "Aci", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H](O)[C@H](N)C1OC(O)(C(=O)O)C[C@H](O)[C@H]1N"},
        "A_ACI": {"name": "Aci", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H](O)[C@H](N)C1O[C@@](O)(C(=O)O)C[C@H](O)[C@H]1N"},
        "B_ACI": {"name": "Aci", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H](O)[C@H](N)C1O[C@](O)(C(=O)O)C[C@H](O)[C@H]1N"},

        # Bacillosamine
        "BAC": {"name": "Bac", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@H]1OC(O)[C@H](N)[C@@H](O)[C@@H]1N"},
        "A_BAC": {"name": "Bac", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@H](O)[C@H](N)[C@@H](O)[C@@H]1N"},
        "B_BAC": {"name": "Bac", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@H]1O[C@@H](O)[C@H](N)[C@@H](O)[C@@H]1N"},

        # L-Glycero-D-Manno-Heptose
        "LDMANHEP": {"name": "LDmanHep", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                     "smiles": "OC[C@H](O)C1OC(O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "A_LDMANHEP": {"name": "LDmanHep", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                       "smiles": "OC[C@H](O)C1O[C@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "B_LDMANHEP": {"name": "LDmanHep", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                       "smiles": "OC[C@H](O)C1O[C@@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},

        # 2-Keto-3-Deoxy-D-Mannooctanoic acid
        "KDO": {"name": "Kdo", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "O=C(O)C1(O)C[C@@H](O)[C@@H](O)C([C@H](O)CO)O1"},
        "A_KDO": {"name": "Kdo", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O=C(O)[C@@]1(O)C[C@@H](O)[C@@H](O)C([C@H](O)CO)O1"},
        "B_KDO": {"name": "Kdo", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O=C(O)[C@@]1(O)C[C@@H](O)[C@@H](O)C([C@H](O)CO)O1"},

        # 3-Deoxy-D-Lyxo-Heptopyran-2-ularic acid
        "DHA": {"name": "Dha", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "O=C(O)[C@H]1OC(O)(C(=O)O)C[C@@H](O)[C@H]1O"},
        "A_DHA": {"name": "Dha", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O=C(O)[C@H]1O[C@@](O)(C(=O)O)C[C@@H](O)[C@H]1O"},
        "B_DHA": {"name": "Dha", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "O=C(O)[C@H]1O[C@](O)(C(=O)O)C[C@@H](O)[C@H]1O"},

        # D-Glycero-D-Manno-Heptose
        "DDMANHEP": {"name": "DDmanHep", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                     "smiles": "OC[C@@H](O)C1OC(O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "A_DDMANHEP": {"name": "DDmanHep", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                       "smiles": "OC[C@@H](O)C1O[C@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "B_DDMANHEP": {"name": "DDmanHep", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                       "smiles": "OC[C@@H](O)C1O[C@@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},

        # Muramic acid
        "MUR": {"name": "Mur", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "C[C@@H](O[C@@H]1[C@@H](N)C(O)O[C@H](CO)[C@H]1O)C(=O)O"},
        "A_MUR": {"name": "Mur", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H](O[C@@H]1[C@@H](N)[C@@H](O)O[C@H](CO)[C@H]1O)C(=O)O"},
        "B_MUR": {"name": "Mur", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "C[C@@H](O[C@@H]1[C@@H](N)[C@H](O)O[C@H](CO)[C@H]1O)C(=O)O"},

        # Apiose
        "API": {"name": "Api", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@@]1(O)COC(O)[C@@H]1O"},
        "A_API": {"name": "Api", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@]1(O)CO[C@H](O)[C@@H]1O"},
        "B_API": {"name": "Api", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@]1(O)CO[C@@H](O)[C@@H]1O"},

        # Fructose
        "FRU": {"name": "Fru", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OCC1(O)OC[C@@H](O)[C@@H](O)[C@@H]1O"},
        "A_FRU": {"name": "Fru", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@]1(O)OC[C@@H](O)[C@@H](O)[C@@H]1O"},
        "B_FRU": {"name": "Fru", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@]1(O)OC[C@@H](O)[C@@H](O)[C@@H]1O"},

        # Tagatose
        "TAG": {"name": "Tag", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OCC1(O)OC[C@@H](O)[C@H](O)[C@@H]1O"},
        "A_TAG": {"name": "Tag", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@]1(O)OC[C@@H](O)[C@H](O)[C@@H]1O"},
        "B_TAG": {"name": "Tag", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@]1(O)OC[C@@H](O)[C@H](O)[C@@H]1O"},

        # Sorbose
        "SOR": {"name": "Sor", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                "smiles": "OCC1(O)OC[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_SOR": {"name": "Sor", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@]1(O)OC[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_SOR": {"name": "Sor", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@]1(O)OC[C@H](O)[C@@H](O)[C@@H]1O"},

        # Psicose
        "PSI": {"name": "Psi", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "OCC1(O)OC[C@@H](O)[C@@H](O)[C@H]1O"},
        "A_PSI": {"name": "Psi", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@]1(O)OC[C@@H](O)[C@@H](O)[C@H]1O"},
        "B_PSI": {"name": "Psi", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                  "smiles": "OC[C@@]1(O)OC[C@@H](O)[C@@H](O)[C@H]1O"},
    }

    def __contains__(self, item):
        """
        Check if an item is part of this factory and can be returned.

        Args:
            item (str): Code of a monomer to be checked

        Returns:
            True if the item is included in the current version of this package
        """
        return item.upper() in PyranoseFactory.__monomers.keys()

    @staticmethod
    def keys():
        """
        Get all monomers that are included in this package by their extended name.

        Returns:
            Set of names for all monomers, their available derivatives and configurations (alpha/beta/undefined)
        """
        return PyranoseFactory.__monomers.keys()

    @staticmethod
    def monomers():
        """
        Get the names of all monomers in this package ignoring the alpha/beta conformations.

        Returns:
            List, sorted from long to short, of all monomer names in upper case
        """
        return list(set(x.split("_")[-1] for x in PyranoseFactory.__monomers.keys()))

    @staticmethod
    def __getitem__(item):
        """
        Get an instance of a monomer from this factory.

        Args:
            item (str): name of the query monomer

        Returns:
            Directory containing all necessary information to initialize a monomer implementation
        """
        return PyranoseFactory.__monomers[item.upper()]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glyles', 'glyles.glycans', 'glyles.glycans.factory', 'glyles.grammar']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime>=4.9.3,<5.0.0',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.21.4,<2.0.0',
 'pydot>=1.4.2,<2.0.0',
 'rdkit-pypi>=2021.9.2,<2022.0.0']

setup_kwargs = {
    'name': 'glyles',
    'version': '0.2.0',
    'description': 'A tool to convert IUPAC representation of glycans into SMILES strings',
    'long_description': '# GlyLES\n\nA tool to convert IUPAC representation of Glycans into SMILES representation. This repo is still in the development \nphase; so, feel free to report any errors or issues.\n\n## Specification and (current) Limitations\n\nThe exact specification, we\'re referring to when talking about "IUPAC representations of glycan", is given in the \n"Notes" section of this [website](https://www.ncbi.nlm.nih.gov/glycans/snfg.html). But as this package is still in the \ndevelopment phase, not everything of the specification is implemented yet (especially not all monomers and side chains \nyou can attach to monomers).\n\nThis implementation currently only works for glycans that fulfil certain properties:\n\n* Linkages have to be explicit, i.e. `(a1-4)`\n* The structure of the glycan can be represented as tree of the monomers with maximal branching factor 2.\n* All root monomers (e.g. Glc, but not GlcNAc) from this [website](https://www.ncbi.nlm.nih.gov/glycans/snfg.html)\n* Some modifications can be added to the monomers, please see the [README](glyles/grammar/README.md) in the grammar\nfolder for more information on this. \n\n## Installation\n\nSo far, this package can only be downloaded from the python package index. So the installation with `pip` is very easy.\nJust type\n\n``````shell\npip install glyles\n``````\n\nand you\'re ready to use it as described below. Use \n\n````shell\npip install --upgrade glyles\n````\n\nto upgrade the glyles package to the most recent version.\n\n## Workflow\n\nConvert the IUPAC into a SMILES representation using the handy `convert` method\n\n``````python\nfrom glyles.converter import convert\n\nconvert(glycan="Man(a1-2)Man", output_file="./test.txt")\n``````\n\nYou can also use the `convert_generator` method to get generator over all SMILES:\n\n``````python\nfrom glyles.converter import convert_generator\n\nfor smiles in convert_generator(glycan_list=["Man(a1-2)Man a", "Man(a1-2)Man b"]):\n    print(smiles)\n``````\n\nIn general, the `convert` and `convert_generator` methods supports the same for types of input. The samples are shown\nfor `convert` but it\'s the same for `convert_generator`.\n\n* single glycan, e.g. `convert(glycan="Man(a1-2)Man)"`,\n* a list of glycans, e.g. `convert(glycan_list=["Man(a1-2)Man a", "Man(a1-2)Man"])`, and\n* a file of glycans, e.g. `convert(glycan_file="./glycans.txt")`. Here its important that the file many only contain one\n  IUPAC per line.\n* for better runtime one can also provide a generator as input, e.g. `convert(glycan_generator=some_generator)`\n\nThe output for `convert` can be manifold as well. For `convert_generator` there is one output. `convert` supports\n\n* `stdout` when specifying no output-related argument, or\n* returned as list of tuples if `returning=true` is set, or\n* writing to an `output_file`, e.g. `convert(glycan="Man(a1-2)Man", output_file="./out.csv")`.\n\nHere each line of the output will state the input IUPAC and the output SMILES. In case of written output separated with \na comma.\n\n\n## Poetry\n\nTo develop this package I used the poetry package manager (see [here](https://python-poetry.org/) for a detailed\ninstruction). It has basically the same functionality as conda but supports the package management better and also \nsupports distinguishing packages into those that are needed to use the package and those that are needed in the \ndevelopment of the package. In order to enable others to work on this repository, we also publish the exact \nspecifications of out poetry environment.\n',
    'author': 'Roman Joeres',
    'author_email': 'Roman.Joeres@helmholtz-hzi.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import numpy as np
from rdkit.Chem import MolFromSmiles, MolToSmiles, GetAdjacencyMatrix
from rdkit.Chem.rdchem import Atom, EditableMol, BondType

from glyles.glycans.monomer import Monomer
from glyles.glycans.utils import UnreachableError, Tree
from glyles.grammar.GlycanLexer import GlycanLexer


class RDKitMonomer(Monomer):

    class Reactor:
        """
        Class with access to protected classes fields managing the modifications of a root monomer.
        """

        def __init__(self, monomer):
            """
            Initialize the reactor with the monomer to modify.

            Args:
                monomer (RDKitMonomer): monomer to be modified
            """
            self.monomer = monomer
            if self.monomer._structure is None:
                self.monomer._get_structure()

        def react(self, names, types):
            """
            Manage the parsed modifications and apply them in turn.

            Args:
                names (List[str]): name (string representation) of the modification
                types (List[int]): Type of the parsed stings based on GlycanLexer.TYPE

            Returns:
                Modified monomer
            """
            for name, t in zip(names, types):
                if t != GlycanLexer.MOD:
                    continue
                if len(name) == 1:
                    if name[0] == "N":
                        self.set_nitrogen()
                    if name[0] == "A":
                        self.make_acid()
                elif len(name) == 2:
                    if name[0].isdigit():
                        if name[1] == "d":  # ?d
                            raise NotImplementedError("Deoxygenation not implemented yet")
                        elif name[1] == "S":
                            self.add_sulfur(int(name[0]))
                        elif name[1] == "P":
                            self.add_phosphate(int(name[0]))
                        elif name[1] == "N":
                            self.set_nitrogen(position=int(name[0]))
                    else:
                        if name == "Ac":
                            self.add_acid(position=5)
                elif len(name) == 3:
                    if name == "NAc":
                        self.add_acid(pos=self.set_nitrogen())
                    elif name[0].isdigit():
                        if name.endswith("Me"):
                            self.add_methyl(position=int(name[0]))
                        elif name.endswith("Ac"):
                            self.add_acid(position=int(name[0]))
                elif len(name) == 7:
                    if name.endswith("Me-"):
                        self.add_methyl(position=int(name[0]))
                    if name.endswith("Ac-"):
                        self.add_acid(position=int(name[0]))

            return self.monomer

        def add_sulfur(self, position):
            """
            Add a SO3- group at the oxygen of the specified position.
            Example: Gal -> Gal3S

            Args:
                position (int): id of the carbon where to add the SO3- group to the bound oxygen

            Returns:
                Nothing
            """
            pos = self.monomer._find_oxygen(position)
            emol = EditableMol(self.monomer._structure)

            s_id = EditableMol.AddAtom(emol, Atom(16))
            o1_id = EditableMol.AddAtom(emol, Atom(8))
            o2_id = EditableMol.AddAtom(emol, Atom(8))
            o3_id = EditableMol.AddAtom(emol, Atom(8))
            EditableMol.AddBond(emol, s_id, o1_id, order=BondType.DOUBLE)
            EditableMol.AddBond(emol, s_id, o2_id, order=BondType.SINGLE)
            EditableMol.AddBond(emol, s_id, o3_id, order=BondType.DOUBLE)
            EditableMol.AddBond(emol, pos, s_id, order=BondType.SINGLE)

            self.monomer._structure = emol.GetMol()

            new_x, new_adj = self._extend_matrices(4)
            new_x[s_id:, 0] = [16, 8, 8, 8]
            self.monomer._x = new_x

            new_adj[s_id, o1_id] = 1
            new_adj[s_id, o2_id] = 1
            new_adj[s_id, o3_id] = 1
            new_adj[s_id, pos] = 1
            new_adj[o1_id, s_id] = 1
            new_adj[o2_id, s_id] = 1
            new_adj[o3_id, s_id] = 1
            new_adj[pos, s_id] = 1

            self.monomer._adjacency = new_adj

        def add_phosphate(self, position):
            """
            Add a PO3 group at the oxygen of the specified position.
            Example: Gal -> Gal3P

            Args:
                position (int): id of the carbon where to add the PO3 group to the bound oxygen

            Returns:
                Nothing
            """
            pos = self.monomer._find_oxygen(position)
            emol = EditableMol(self.monomer._structure)

            s_id = EditableMol.AddAtom(emol, Atom(15))
            o1_id = EditableMol.AddAtom(emol, Atom(8))
            o2_id = EditableMol.AddAtom(emol, Atom(8))
            o3_id = EditableMol.AddAtom(emol, Atom(8))
            EditableMol.AddBond(emol, s_id, o1_id, order=BondType.SINGLE)
            EditableMol.AddBond(emol, s_id, o2_id, order=BondType.SINGLE)
            EditableMol.AddBond(emol, s_id, o3_id, order=BondType.DOUBLE)
            EditableMol.AddBond(emol, pos, s_id, order=BondType.SINGLE)

            self.monomer._structure = emol.GetMol()

            new_x, new_adj = self._extend_matrices(4)
            new_x[s_id:, 0] = [15, 8, 8, 8]
            self.monomer._x = new_x

            new_adj[s_id, o1_id] = 1
            new_adj[s_id, o2_id] = 1
            new_adj[s_id, o3_id] = 1
            new_adj[s_id, pos] = 1
            new_adj[o1_id, s_id] = 1
            new_adj[o2_id, s_id] = 1
            new_adj[o3_id, s_id] = 1
            new_adj[pos, s_id] = 1

            self.monomer._adjacency = new_adj

        def add_acid(self, position=None, pos=None):
            """
            Add an acid group to a specific position. Here the position can be provided either as the C-index
            (position) or as the rdkit id of the atom where to append the acid group (implemented for NAc). Exactly one
            of both arguments must be provided.
            Example: GalN -> GalNAc or Gal -> Gal5Ac

            Args:
                position (int): index of the c-atom where to append the acid group
                pos (int): rdkit id of the atom where to append the acid group

            Returns:
                Nothing
            """
            if (position is None) == (pos is None):
                raise ValueError()

            if position is not None:
                pos = self.monomer._find_oxygen(position, check_for=[8, 7])

            emol = EditableMol(self.monomer._structure)

            c1_id = EditableMol.AddAtom(emol, Atom(6))
            c2_id = EditableMol.AddAtom(emol, Atom(6))
            o1_id = EditableMol.AddAtom(emol, Atom(8))
            EditableMol.AddBond(emol, c1_id, pos, order=BondType.SINGLE)
            EditableMol.AddBond(emol, c2_id, c1_id, order=BondType.SINGLE)
            EditableMol.AddBond(emol, o1_id, c1_id, order=BondType.DOUBLE)

            self.monomer._structure = emol.GetMol()

            new_x, new_adj = self._extend_matrices(3)
            new_x[c1_id:, 0] = [6, 6, 8]
            self.monomer._x = new_x

            new_adj[c1_id, c2_id] = 1
            new_adj[c1_id, o1_id] = 1
            new_adj[c1_id, pos] = 1
            new_adj[c2_id, c1_id] = 1
            new_adj[o1_id, c1_id] = 1
            new_adj[pos, c1_id] = 1

            self.monomer._adjacency = new_adj

        def add_methyl(self, position):
            """

            Args:
                position:

            Returns:

            """
            pos = self.monomer._find_oxygen(position)

            emol = EditableMol(self.monomer._structure)

            c_id = EditableMol.AddAtom(emol, Atom(6))
            EditableMol.AddBond(emol, c_id, pos, order=BondType.SINGLE)

            self.monomer._structure = emol.GetMol()

            new_x, new_adj = self._extend_matrices(1)
            new_x[c_id, 0] = 6
            self.monomer._x = new_x

            new_adj[c_id, pos] = 1
            new_adj[pos, c_id] = 1

            self.monomer._adjacency = new_adj

        def set_nitrogen(self, position=2):
            """
            Change an oxygen to a nitrogen.
            Example: Gal -> GalN

            Args:
                position (int): position of a carbon atom at which the bound o atom should be replaced by N

            Returns:
                rdkit id of the atom that is now a nitrogen
            """
            pos = self.monomer._find_oxygen(position, check_for=[8, 7])
            self.monomer._structure.GetAtomWithIdx(pos).SetAtomicNum(7)
            self.monomer._x[pos, 0] = 7
            return pos

        def make_acid(self):
            """

            Returns:

            """
            emol = EditableMol(self.monomer._structure)

            c_id = int(np.argwhere(self.monomer._x[:, 1] == 6))
            o_id = EditableMol.AddAtom(emol, Atom(8))
            EditableMol.AddBond(emol, o_id, c_id, order=BondType.DOUBLE)

            self.monomer._structure = emol.GetMol()

            new_x, new_adj = self._extend_matrices(1)
            new_x[o_id, 0] = 8
            self.monomer._x = new_x

            new_adj[c_id, o_id] = 1
            new_adj[o_id, c_id] = 1
            self.monomer._adjacency = new_adj

        def _extend_matrices(self, count):
            """

            Args:
                count:

            Returns:

            """
            tmp_x = np.zeros((self.monomer._x.shape[0] + count, self.monomer._x.shape[1]))
            tmp_x[:self.monomer._x.shape[0], :] = self.monomer._x
            tmp_adj = np.zeros((self.monomer._adjacency.shape[0] + count, self.monomer._adjacency.shape[1] + count))
            tmp_adj[:self.monomer._adjacency.shape[0], :self.monomer._adjacency.shape[1]] = self.monomer._adjacency
            return tmp_x, tmp_adj

    def __init__(self, origin=None, **kwargs):
        """
        Initialize the monomer using the super method. Additionally, some fields are initialized to describe the
        structure of the monomer according to the specification of the monomer-parent class

        Args:
            origin (Monomer): Other monomer to use to initialize this object
            **kwargs: arguments to initialize monomer if object is None. Must include name, SMILES, and config
        """
        super(RDKitMonomer, self).__init__(origin, **kwargs)
        if isinstance(origin, RDKitMonomer):
            self._adjacency = origin.get_adjacency()
            self._ring_info = origin.get_ring_info()
            self._x = origin.get_features()
        else:
            self._adjacency = None
            self._ring_info = None
            self._x = None
            self._get_structure()

    def alpha(self, factory):
        """
        Return this monosaccharide in its alpha conformation.

        Args:
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            Monomer in alpha conformation
        """
        recipe = [(v, t) for v, t in self._recipe if t != GlycanLexer.TYPE]
        recipe.append(('a', GlycanLexer.TYPE))
        return RDKitMonomer(factory.create(recipe))

    def beta(self, factory):
        """
        Return this monosaccharide in its beta conformation.

        Args:
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            Monomer in beta conformation
        """
        recipe = [(v, t) for v, t in self._recipe if t != GlycanLexer.TYPE]
        recipe.append(('b', GlycanLexer.TYPE))
        return RDKitMonomer(factory.create(recipe))

    def undefined(self, factory):
        """
        Return this monosaccharide in undefined conformation, the first carbon ring-atom will have unspecified.
        chirality.

        Args:
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            Monomer in undefined conformation
        """
        recipe = [(v, t) for v, t in self._recipe if t != GlycanLexer.TYPE]
        return RDKitMonomer(factory.create(recipe))

    def get_adjacency(self):
        """
        Get the adjacency matrix of the atoms in this monomer.

        Returns:
            Adjacency matrix of all non-hydrogen atoms in this monomer
        """
        if self._adjacency is None:
            self._get_structure()
        return self._adjacency

    def get_ring_info(self):
        """
        Get information of all atom-ids in the rings of this monomer.

        Returns:
            Tuple of tuples with the atom-ids from rdkit in the monomer
        """
        if self._ring_info is None:
            self._get_structure()
        return self._ring_info

    def get_features(self):
        """
        Get a feature matrix from this monomer. The features contain information about atom type, ids according to the
        specification in monomer.py, and ring memberships.

        Returns:
            A numpy array of shape Nx3 containing the extracted features for all atoms in this molecule
        """
        if self._x is None:
            self._get_structure()
        return self._x

    def get_dummy_atoms(self):
        """
        Specify some dummy atoms that are used to mark oxygen atoms that will participate in bindings between glycans.
        Here, the atoms will be replaced by instances of the atom enum that are used to define the type of the atom in
        the nodes of the networkx representation of the monomer molecules.
        TODO: Extend this to N- and C-glycosidic bonds

        Returns:
            Two lists:
                * one with atoms that are given as the "atom" argument to the mark-method to replace the oxygen atoms
                * the string representation of the atoms from above, i.e. how the atoms above will be represented in a
                  SMILES string
        """
        return [34, 52, 84], ["[SeH]", "[TeH]", "[PoH]"]

    def root_atom_id(self, binding_c_id):
        """
        Get ID of atom that will bind the parent monomer in the glycan. This ID will be given as root argument to
        the to_smiles method.

        Args:
            binding_c_id (int): Integer at which c-position this monomer binds its parent

        Returns:
            id of the atom that binds to the parent, -1 if the root cannot be found
        """
        return self._find_oxygen(binding_c_id)

    def mark(self, position, atom):
        """
        Mark the oxygen atom linked to the carbon atom at the given position ready to participate in the bounding.
        Marking here works based on replacing the oxygen-group bound to the carbon atom at the given position with the
        atom enum instance also provided in the arguments.

        Args:
            position (int): id of the carbon atom whose oxygen atom will from the binding
            atom (int): atom to replace the binding oxygen with

        Returns:
            Nothing
        """
        idx = self._find_oxygen(position)
        self._get_structure().GetAtomWithIdx(idx).SetAtomicNum(atom)
        self._x[idx, 0] = atom

    def to_smiles(self, root, ring_index):
        """
        Convert this monomer into a SMILES string representation.
        Use the implementation of the SMILES algorithm fitted to the needs of glycans.

        Args:
            root (int): index of the root atom
            ring_index (int): index of the rings in the atom

        Returns:
            SMILES string representation of this molecule
        """
        smiles = MolToSmiles(self._get_structure(), rootedAtAtom=root)
        smiles.replace("At", "O-")
        return "".join([(str(int(c) + ring_index) if c.isdigit() else c) for c in smiles])

    def react(self, names, types):
        """
        Override the method to the call of the reactor to modify this monomer.

        Args:
            names (List[str]): name (string representation) of the modification
            types (List[int]): Type of the parsed stings based on GlycanLexer.TYPE

        Returns:
            New monomer with the altered structure
        """
        return RDKitMonomer.Reactor(self).react(names, types)

    def _get_structure(self):
        """
        Compute and save the structure of this glycan.

        Returns:
            rdkit molecule representing the structure of the glycan as a graph of its non-hydrogen atoms.
        """
        if self._structure is None:
            # read the structure from the SMILES string
            self._structure = MolFromSmiles(self._smiles)

            # extract some further information from the molecule to not operate always on the molecule
            self._adjacency = GetAdjacencyMatrix(self._structure)
            self._ring_info = self._structure.GetRingInfo().AtomRings()
            self._x = np.zeros((self._adjacency.shape[0], 3))

            c_atoms, ringo = [], -1
            # extract some information form the molecule
            for i in range(self._adjacency.shape[0]):
                atom = self._structure.GetAtomWithIdx(i)

                # store the atom type
                self._x[i, 0] = atom.GetAtomicNum()
                if self._x[i, 0] == 6 and i in self._ring_info[0]:
                    c_atoms.append(int(i))

                # if the atom is part of any ring, store the number of that ring
                for r in range(len(self._ring_info)):
                    if i in self._ring_info[r]:
                        self._x[i, 2] = r + 1

                # identify the oxygen atom in the main ring and set its id to 10
                if self._x[i, 2] == 1 and self._x[i, 0] == 8:
                    self._x[i, 1] = 10
                    ringo = i

            self._enumerate_c_atoms(c_atoms, ringo)
        return self._structure

    def _equidistant(self, start, end):
        """
        Decider for C1 in case the previous splitting rules were all tied. This currently only fires for
        Fruf, Tagf, Sorf, Psif

        Args:
            start (int): id of the first candidate for C1 atom
            end (int): id of the second candidate for C1 atom

        Returns:
            Bool indicating that the start id is the C1 atom
        """
        c_start_candidates = np.argwhere((self._adjacency[start, :] == 1) &
                                         (self._x[:, 0] == 6) & (self._x[:, 2] == 1)).squeeze()
        c_end_candidates = np.argwhere((self._adjacency[end, :] == 1) &
                                       (self._x[:, 0] == 6) & (self._x[:, 2] == 1)).squeeze()
        if c_start_candidates.size == 1 and c_end_candidates.size == 1:
            start_ring_c = int(c_start_candidates)
            end_ring_c = int(c_end_candidates)

            start_ring_c_o_candidates = np.argwhere((self._adjacency[start_ring_c, :] == 1) &
                                                    (self._x[:, 0] == 8) & (self._x[:, 2] != 1)).squeeze()
            end_ring_c_o_candidates = np.argwhere((self._adjacency[end_ring_c, :] == 1) &
                                                  (self._x[:, 0] == 8) & (self._x[:, 2] != 1)).squeeze()

            if start_ring_c_o_candidates.size == 1 and end_ring_c_o_candidates.size == 1:
                raise UnreachableError("C1 atom cannot be detected")
            elif start_ring_c_o_candidates.size == 1:
                return True
        elif c_start_candidates.size == 1:
            return True
        return False

    def _evaluate_distance(self, start, end, ringo):
        """
        Try to decide on C1 based on their distance to the oxygen in the ring

        Args:
            start (int): id of the first candidate for C1 atom
            end (int): id of the second candidate for C1 atom
            ringo (int): index of the oxygen atom in the ring

        Returns:
            Bool indicating that the start id is the C1 atom
        """
        adj = self._adjacency.copy()

        # as we have an adjacency matrix, multiply it with itself until one of the fields is non-zero
        while adj[start, ringo] == 0 and adj[end, ringo] == 0:
            adj = adj @ self._adjacency

        # if both fields are non-zero, we cannot decide here and have to go further
        if adj[start, ringo] > 0 and adj[end, ringo] > 0:
            return self._equidistant(start, end)
        elif adj[start, ringo] > 0:
            return True
        return False

    def _enumerate_c_atoms(self, c_atoms, ringo):
        """
        Enumerate all carbon atoms starting from the first one

        Args:
            c_atoms List[int]: List of all ids of C atoms in the ring
            ringo (int): id of the oxygen atom in the ring of the monomer

        Returns:
            Nothing
        """
        # create a tree of all carbon atoms directly connected to the main ring of the monomer
        c_tree = Tree()
        stack = [(-1, c_atoms[0])]
        while len(stack) != 0:
            p_id, c_id = stack[-1]
            stack = stack[:-1]
            c_tree.add_node(c_id, p_id)

            children = np.argwhere(self._adjacency[c_id] & (self._x[:, 0] == 6))
            for c in children:
                if int(c) not in c_tree.nodes:
                    stack.append((c_id, int(c)))

        # find the deepest node and rehang the tree to this node
        deepest_id, _ = c_tree.deepest_node()
        c_tree = c_tree.rehang_tree(deepest_id)
        longest_c_chain = c_tree.longest_chain()

        # now the two C1 candidates can be found at the ends of the longest chain
        start, end = longest_c_chain[0], longest_c_chain[-1]

        # check conditions
        start_o_conn = np.argwhere((self._adjacency[start, :] == 1) & (self._x[:, 0] == 8) &
                                   (self._x[:, 2] != 1)).squeeze().size > 0
        end_o_conn = np.argwhere((self._adjacency[end, :] == 1) & (self._x[:, 0] == 8) &
                                 (self._x[:, 2] != 1)).squeeze().size > 0

        # decide on c1
        if start_o_conn and end_o_conn:
            if not self._evaluate_distance(start, end, ringo):
                longest_c_chain = reversed(longest_c_chain)
        elif end_o_conn:
            longest_c_chain = reversed(longest_c_chain)

        # enumerate along chain
        c_count = 0
        for c in longest_c_chain:
            c_count += 1
            self._x[c, 1] = c_count

    def _find_oxygen(self, binding_c_id, check_for=None):
        """
        Find the oxygen atom that binds to the carbon atom with the provided id. The returned id may not refer to an
        oxygen atom in the ring of the monomer as this cannot bind anything. This method will report the atom id of the
        first atom type that fulfils the requirements of this method, i.e. the binding atom has to be the only one of
        its type, not in the main-ring, and connected to the given carbon atom.

        Args:
            binding_c_id (int): id of the carbon atom that participates in a binding, and we need to find the oxygen
                from
            check_for (List[int]): List if atom types as numbered in PSE to check for binding the above given carbon

        Returns:
            id referring to the atom binding the provided carbon atom and may participate in a glycan-binding.
        """
        # by default only check for binding oxygen
        if check_for is None:
            check_for = [8]

        # first find the rdkit id of the carbon atom that should bind to something
        position = np.argwhere(self._x[:, 1] == binding_c_id).squeeze()

        for check in check_for:
            # then find the candidates. There should be exactly one element in the resulting array
            candidates = np.argwhere((self._adjacency[position, :] == 1) &
                                     (self._x[:, 0] == check) & (self._x[:, 2] != 1)).squeeze()
            if candidates.size == 1:
                return int(candidates)

        raise ValueError(f"Multiple (or no) options for oxygen (or other atom type) found.")



# Generated from Glycan.g4 by ANTLR 4.9
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\r")
        buf.write("Q\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\3\2\3\2\3\2\3\2\3\2")
        buf.write("\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3")
        buf.write("\2\3\2\3\2\3\2\5\2!\n\2\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3")
        buf.write("\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\5\3\65\n\3")
        buf.write("\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\5\4B\n\4")
        buf.write("\3\5\7\5E\n\5\f\5\16\5H\13\5\3\5\3\5\7\5L\n\5\f\5\16\5")
        buf.write("O\13\5\3\5\2\2\6\2\4\6\b\2\2\2W\2 \3\2\2\2\4\64\3\2\2")
        buf.write("\2\6A\3\2\2\2\bF\3\2\2\2\n\13\7\3\2\2\13\f\5\4\3\2\f\r")
        buf.write("\5\6\4\2\r\16\7\4\2\2\16\17\7\13\2\2\17\20\7\5\2\2\20")
        buf.write("!\3\2\2\2\21\22\7\3\2\2\22\23\5\4\3\2\23\24\5\6\4\2\24")
        buf.write("\25\7\5\2\2\25!\3\2\2\2\26\27\7\3\2\2\27\30\5\6\4\2\30")
        buf.write("\31\7\4\2\2\31\32\7\13\2\2\32\33\7\5\2\2\33!\3\2\2\2\34")
        buf.write("\35\7\3\2\2\35\36\5\6\4\2\36\37\7\5\2\2\37!\3\2\2\2 \n")
        buf.write("\3\2\2\2 \21\3\2\2\2 \26\3\2\2\2 \34\3\2\2\2!\3\3\2\2")
        buf.write("\2\"#\5\6\4\2#$\7\n\2\2$\65\3\2\2\2%&\5\6\4\2&\'\7\n\2")
        buf.write("\2\'(\5\4\3\2(\65\3\2\2\2)*\7\6\2\2*+\5\4\3\2+,\7\7\2")
        buf.write("\2,\65\3\2\2\2-.\5\6\4\2./\7\n\2\2/\60\7\6\2\2\60\61\5")
        buf.write("\4\3\2\61\62\7\7\2\2\62\63\5\4\3\2\63\65\3\2\2\2\64\"")
        buf.write("\3\2\2\2\64%\3\2\2\2\64)\3\2\2\2\64-\3\2\2\2\65\5\3\2")
        buf.write("\2\2\66\67\5\b\5\2\678\7\f\2\289\7\13\2\29B\3\2\2\2:;")
        buf.write("\5\b\5\2;<\7\13\2\2<B\3\2\2\2=>\5\b\5\2>?\7\f\2\2?B\3")
        buf.write("\2\2\2@B\5\b\5\2A\66\3\2\2\2A:\3\2\2\2A=\3\2\2\2A@\3\2")
        buf.write("\2\2B\7\3\2\2\2CE\7\t\2\2DC\3\2\2\2EH\3\2\2\2FD\3\2\2")
        buf.write("\2FG\3\2\2\2GI\3\2\2\2HF\3\2\2\2IM\7\b\2\2JL\7\t\2\2K")
        buf.write("J\3\2\2\2LO\3\2\2\2MK\3\2\2\2MN\3\2\2\2N\t\3\2\2\2OM\3")
        buf.write("\2\2\2\7 \64AFM")
        return buf.getvalue()


class GlycanParser ( Parser ):

    grammarFileName = "Glycan.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'{'", "' '", "'}'", "'['", "']'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "SAC", "MOD", "CON", "TYPE", 
                      "RING", "NUM" ]

    RULE_start = 0
    RULE_branch = 1
    RULE_glycan = 2
    RULE_deriv = 3

    ruleNames =  [ "start", "branch", "glycan", "deriv" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    SAC=6
    MOD=7
    CON=8
    TYPE=9
    RING=10
    NUM=11

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.9")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StartContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def branch(self):
            return self.getTypedRuleContext(GlycanParser.BranchContext,0)


        def glycan(self):
            return self.getTypedRuleContext(GlycanParser.GlycanContext,0)


        def TYPE(self):
            return self.getToken(GlycanParser.TYPE, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_start

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStart" ):
                listener.enterStart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStart" ):
                listener.exitStart(self)




    def start(self):

        localctx = GlycanParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        try:
            self.state = 30
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 8
                self.match(GlycanParser.T__0)
                self.state = 9
                self.branch()
                self.state = 10
                self.glycan()
                self.state = 11
                self.match(GlycanParser.T__1)
                self.state = 12
                self.match(GlycanParser.TYPE)
                self.state = 13
                self.match(GlycanParser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 15
                self.match(GlycanParser.T__0)
                self.state = 16
                self.branch()
                self.state = 17
                self.glycan()
                self.state = 18
                self.match(GlycanParser.T__2)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 20
                self.match(GlycanParser.T__0)
                self.state = 21
                self.glycan()
                self.state = 22
                self.match(GlycanParser.T__1)
                self.state = 23
                self.match(GlycanParser.TYPE)
                self.state = 24
                self.match(GlycanParser.T__2)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 26
                self.match(GlycanParser.T__0)
                self.state = 27
                self.glycan()
                self.state = 28
                self.match(GlycanParser.T__2)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BranchContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def glycan(self):
            return self.getTypedRuleContext(GlycanParser.GlycanContext,0)


        def CON(self):
            return self.getToken(GlycanParser.CON, 0)

        def branch(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.BranchContext)
            else:
                return self.getTypedRuleContext(GlycanParser.BranchContext,i)


        def getRuleIndex(self):
            return GlycanParser.RULE_branch

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBranch" ):
                listener.enterBranch(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBranch" ):
                listener.exitBranch(self)




    def branch(self):

        localctx = GlycanParser.BranchContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_branch)
        try:
            self.state = 50
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 32
                self.glycan()
                self.state = 33
                self.match(GlycanParser.CON)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 35
                self.glycan()
                self.state = 36
                self.match(GlycanParser.CON)
                self.state = 37
                self.branch()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 39
                self.match(GlycanParser.T__3)
                self.state = 40
                self.branch()
                self.state = 41
                self.match(GlycanParser.T__4)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 43
                self.glycan()
                self.state = 44
                self.match(GlycanParser.CON)
                self.state = 45
                self.match(GlycanParser.T__3)
                self.state = 46
                self.branch()
                self.state = 47
                self.match(GlycanParser.T__4)
                self.state = 48
                self.branch()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class GlycanContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def deriv(self):
            return self.getTypedRuleContext(GlycanParser.DerivContext,0)


        def RING(self):
            return self.getToken(GlycanParser.RING, 0)

        def TYPE(self):
            return self.getToken(GlycanParser.TYPE, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_glycan

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterGlycan" ):
                listener.enterGlycan(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitGlycan" ):
                listener.exitGlycan(self)




    def glycan(self):

        localctx = GlycanParser.GlycanContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_glycan)
        try:
            self.state = 63
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 52
                self.deriv()
                self.state = 53
                self.match(GlycanParser.RING)
                self.state = 54
                self.match(GlycanParser.TYPE)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 56
                self.deriv()
                self.state = 57
                self.match(GlycanParser.TYPE)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 59
                self.deriv()
                self.state = 60
                self.match(GlycanParser.RING)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 62
                self.deriv()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DerivContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def SAC(self):
            return self.getToken(GlycanParser.SAC, 0)

        def MOD(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.MOD)
            else:
                return self.getToken(GlycanParser.MOD, i)

        def getRuleIndex(self):
            return GlycanParser.RULE_deriv

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDeriv" ):
                listener.enterDeriv(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDeriv" ):
                listener.exitDeriv(self)




    def deriv(self):

        localctx = GlycanParser.DerivContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_deriv)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 68
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==GlycanParser.MOD:
                self.state = 65
                self.match(GlycanParser.MOD)
                self.state = 70
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 71
            self.match(GlycanParser.SAC)
            self.state = 75
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==GlycanParser.MOD:
                self.state = 72
                self.match(GlycanParser.MOD)
                self.state = 77
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





